@@ -12,6 +12,8 @@ const (
 	fecHeaderSizePlus2 = fecHeaderSize + 2 // plus 2B data size
 	typeData           = 0xf1
 	typeParity         = 0xf2
+	fecExpire          = 60000
+	rxFECMulti         = 3 // FEC keeps rxFECMulti* (dataShard+parityShard) ordered packets in memory
 )
 
 // fecPacket is a decoded FEC packet
@@ -21,13 +23,19 @@ func (bts fecPacket) seqid() uint32 { return binary.LittleEndian.Uint32(bts) }
 func (bts fecPacket) flag() uint16  { return binary.LittleEndian.Uint16(bts[4:]) }
 func (bts fecPacket) data() []byte  { return bts[6:] }
 
+// fecElement has auxcilliary time field
+type fecElement struct {
+	fecPacket
+	ts uint32
+}
+
 // fecDecoder for decoding incoming packets
 type fecDecoder struct {
 	rxlimit      int // queue size limit
 	dataShards   int
 	parityShards int
 	shardSize    int
-	rx           []fecPacket // ordered receive queue
+	rx           []fecElement // ordered receive queue
 
 	// caches
 	decodeCache [][]byte
@@ -38,21 +46,21 @@ type fecDecoder struct {
 
 	// RS decoder
 	codec reedsolomon.Encoder
+
+	// auto tune fec parameter
+	autoTune autoTune
 }
 
-func newFECDecoder(rxlimit, dataShards, parityShards int) *fecDecoder {
+func newFECDecoder(dataShards, parityShards int) *fecDecoder {
 	if dataShards <= 0 || parityShards <= 0 {
 		return nil
 	}
-	if rxlimit < dataShards+parityShards {
-		return nil
-	}
 
 	dec := new(fecDecoder)
-	dec.rxlimit = rxlimit
 	dec.dataShards = dataShards
 	dec.parityShards = parityShards
 	dec.shardSize = dataShards + parityShards
+	dec.rxlimit = rxFECMulti * dec.shardSize
 	codec, err := reedsolomon.New(dataShards, parityShards)
 	if err != nil {
 		return nil
@@ -66,6 +74,49 @@ func newFECDecoder(rxlimit, dataShards, parityShards int) *fecDecoder {
 
 // decode a fec packet
 func (dec *fecDecoder) decode(in fecPacket) (recovered [][]byte) {
+	// sample to auto FEC tuner
+	if in.flag() == typeData {
+		dec.autoTune.Sample(true, in.seqid())
+	} else {
+		dec.autoTune.Sample(false, in.seqid())
+	}
+
+	// check if FEC parameters is out of sync
+	var shouldTune bool
+	if int(in.seqid())%dec.shardSize < dec.dataShards {
+		if in.flag() != typeData { // expect typeData
+			shouldTune = true
+		}
+	} else {
+		if in.flag() != typeParity {
+			shouldTune = true
+		}
+	}
+
+	if shouldTune {
+		autoDS := dec.autoTune.FindPeriod(true)
+		autoPS := dec.autoTune.FindPeriod(false)
+
+		// edges found, we can tune parameters now
+		if autoDS > 0 && autoPS > 0 && autoDS < 256 && autoPS < 256 {
+			// and make sure it's different
+			if autoDS != dec.dataShards || autoPS != dec.parityShards {
+				dec.dataShards = autoDS
+				dec.parityShards = autoPS
+				dec.shardSize = autoDS + autoPS
+				dec.rxlimit = rxFECMulti * dec.shardSize
+				codec, err := reedsolomon.New(autoDS, autoPS)
+				if err != nil {
+					return nil
+				}
+				dec.codec = codec
+				dec.decodeCache = make([][]byte, dec.shardSize)
+				dec.flagCache = make([]bool, dec.shardSize)
+				//log.Println("autotune to :", dec.dataShards, dec.parityShards)
+			}
+		}
+	}
+
 	// insertion
 	n := len(dec.rx) - 1
 	insertIdx := 0
@@ -81,14 +132,15 @@ func (dec *fecDecoder) decode(in fecPacket) (recovered [][]byte) {
 	// make a copy
 	pkt := fecPacket(xmitBuf.Get().([]byte)[:len(in)])
 	copy(pkt, in)
+	elem := fecElement{pkt, currentMs()}
 
 	// insert into ordered rx queue
 	if insertIdx == n+1 {
-		dec.rx = append(dec.rx, pkt)
+		dec.rx = append(dec.rx, elem)
 	} else {
-		dec.rx = append(dec.rx, fecPacket{})
+		dec.rx = append(dec.rx, fecElement{})
 		copy(dec.rx[insertIdx+1:], dec.rx[insertIdx:]) // shift right
-		dec.rx[insertIdx] = pkt
+		dec.rx[insertIdx] = elem
 	}
 
 	// shard range for current packet
@@ -148,7 +200,7 @@ func (dec *fecDecoder) decode(in fecPacket) (recovered [][]byte) {
 					dlen := len(shards[k])
 					shards[k] = shards[k][:maxlen]
 					copy(shards[k][dlen:], dec.zeros)
-				} else {
+				} else if k < dec.dataShards {
 					shards[k] = xmitBuf.Get().([]byte)[:0]
 				}
 			}
@@ -171,13 +223,27 @@ func (dec *fecDecoder) decode(in fecPacket) (recovered [][]byte) {
 		}
 		dec.rx = dec.freeRange(0, 1, dec.rx)
 	}
+
+	// timeout policy
+	current := currentMs()
+	numExpired := 0
+	for k := range dec.rx {
+		if _itimediff(current, dec.rx[k].ts) > fecExpire {
+			numExpired++
+			continue
+		}
+		break
+	}
+	if numExpired > 0 {
+		dec.rx = dec.freeRange(0, numExpired, dec.rx)
+	}
 	return
 }
 
 // free a range of fecPacket
-func (dec *fecDecoder) freeRange(first, n int, q []fecPacket) []fecPacket {
+func (dec *fecDecoder) freeRange(first, n int, q []fecElement) []fecElement {
 	for i := first; i < first+n; i++ { // recycle buffer
-		xmitBuf.Put([]byte(q[i]))
+		xmitBuf.Put([]byte(q[i].fecPacket))
 	}
 
 	if first == 0 && n < cap(q)/2 {
@@ -187,6 +253,13 @@ func (dec *fecDecoder) freeRange(first, n int, q []fecPacket) []fecPacket {
 	return q[:len(q)-n]
 }
 
+// release all segments back to xmitBuf
+func (dec *fecDecoder) release() {
+	if n := len(dec.rx); n > 0 {
+		dec.rx = dec.freeRange(0, n, dec.rx)
+	}
+}
+
 type (
 	// fecEncoder for encoding outgoing packets
 	fecEncoder struct {
